@@ -29,6 +29,7 @@
 #include <X11/Xatom.h>
 #include <unistd.h>
 #include <stdlib.h>
+#include <stdio.h>
 #include <stdarg.h>
 #include <atk/atk.h>
 
@@ -62,10 +63,7 @@ static const AtkMisc *atk_misc = NULL;
  * Returns a 'canonicalized' value for DISPLAY,
  * with the screen number stripped off if present.
  *
- * Not currently used in D-Bus version but may be
- * useful in future if we make use of XAtom.
  */
-#if 0
 static const gchar*
 spi_display_name (void)
 {
@@ -97,7 +95,6 @@ spi_display_name (void)
       }
     return canonical_display_name;
 }
-#endif
 
 /*---------------------------------------------------------------------------*/
 
@@ -107,31 +104,58 @@ spi_display_name (void)
  * may be employed in the future for accessing the registry daemon
  * bus name.
  */
-#if 0
-static gchar *
-spi_atk_bridge_get_registry_ior (void)
+
+static DBusConnection *
+spi_atk_bridge_get_bus (void)
 {
-     Atom AT_SPI_IOR;
+     Atom AT_SPI_BUS;
      Atom actual_type;
+     Display *bridge_display;
      int actual_format;
      unsigned char *data = NULL;  
      unsigned long nitems;
      unsigned long leftover;
-     if (!bridge_display) 
-       bridge_display = XOpenDisplay (spi_display_name ());
 
-     AT_SPI_IOR = XInternAtom (bridge_display, "AT_SPI_IOR", False); 
+     DBusConnection *bus = NULL;
+     DBusError       error;
+
+     bridge_display = XOpenDisplay (spi_display_name ());
+     if (!bridge_display)
+	g_error ("AT_SPI: Could not get the display\n");
+
+     AT_SPI_BUS = XInternAtom (bridge_display, "AT_SPI_BUS", False); 
      XGetWindowProperty(bridge_display, 
                         XDefaultRootWindow (bridge_display),
-                        AT_SPI_IOR, 0L,
+                        AT_SPI_BUS, 0L,
                         (long)BUFSIZ, False,
                         (Atom) 31, &actual_type, &actual_format,
                         &nitems, &leftover, &data);
+
+     dbus_error_init (&error);
+
      if (data == NULL)
-         g_warning (_("AT_SPI_REGISTRY was not started at session startup."));
-     return (gchar *) data;
+     {
+         g_warning ("AT-SPI: Accessibility bus not found - Using session bus.\n");
+         bus = dbus_bus_get (DBUS_BUS_SESSION, &error);
+         if (!bus)
+             g_error ("AT-SPI: Couldn't connect to bus: %s\n", error.message);
+     }
+     else
+     {
+	 bus = dbus_connection_open (data, &error);
+         if (!bus)
+         {
+             g_error ("AT-SPI: Couldn't connect to bus: %s\n", error.message);
+         }
+	 else
+         {
+	     if (!dbus_bus_register (bus, &error))
+	         g_error ("AT-SPI: Couldn't register with bus: %s\n");
+         } 
+     }
+
+     return bus;
 }
-#endif
 
 /*---------------------------------------------------------------------------*/
 
@@ -141,17 +165,21 @@ register_application (SpiAppData *app)
   DBusMessage *message;
   DBusMessageIter iter;
   DBusError error;
-  const char *uname;
+  const char *uname = NULL;
 
   dbus_error_init (&error);
 
   message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
                                           SPI_DBUS_PATH_REGISTRY,
                                           SPI_DBUS_INTERFACE_REGISTRY,
-                                          "registerApplication");
+                                          "RegisterApplication");
   dbus_message_set_no_reply (message, TRUE);
 
   uname = dbus_bus_get_unique_name(app->bus);
+  if (!uname)
+  {
+      g_error ("AT-SPI: Couldn't get unique name for this connection");
+  }
 
   dbus_message_iter_init_append(message, &iter);
   dbus_message_iter_append_basic(&iter, DBUS_TYPE_STRING, &uname);
@@ -174,7 +202,7 @@ deregister_application (SpiAppData *app)
   message = dbus_message_new_method_call (SPI_DBUS_NAME_REGISTRY,
                                           SPI_DBUS_PATH_REGISTRY,
                                           SPI_DBUS_INTERFACE_REGISTRY,
-                                          "deregisterApplication");
+                                          "DeregisterApplication");
   dbus_message_set_no_reply (message, TRUE);
 
   uname = dbus_bus_get_unique_name(app->bus);
@@ -211,8 +239,49 @@ exit_func (void)
 
 /*---------------------------------------------------------------------------*/
 
-static gchar *atspi_dbus_name;
-static gboolean atspi_no_register;
+#ifdef __ATK_PLUG_H__
+static AtkPlugClass *plug_class;
+static AtkSocketClass *socket_class;
+
+static gchar *
+get_plug_id (AtkPlug *plug)
+{
+  const char *uname = dbus_bus_get_unique_name(atk_adaptor_app_data->bus);
+  gchar *path;
+  GString *str = g_string_new (NULL);
+
+  path = atk_dbus_object_to_path (ATK_OBJECT(plug), TRUE);
+  g_string_printf (str, "%s:%s", uname, path);
+  g_free (path);
+  return g_string_free (str, FALSE);
+}
+
+static void
+socket_embed_hook (AtkSocket *socket, gchar *plug_id)
+{
+  AtkObject *accessible = ATK_OBJECT(socket);
+  /* Force registration */
+  gchar *path = atk_dbus_object_to_path (accessible, TRUE);
+  spi_emit_cache_update (accessible, atk_adaptor_app_data->bus);
+  g_free (path);
+}
+
+static void
+install_plug_hooks ()
+{
+  gpointer data;
+  
+  data = g_type_class_ref (ATK_TYPE_PLUG);
+  plug_class = ATK_PLUG_CLASS (data);
+  data = g_type_class_ref (ATK_TYPE_SOCKET);
+  socket_class = ATK_SOCKET_CLASS (data);
+  plug_class->get_object_id = get_plug_id;
+  socket_class->embed = socket_embed_hook;
+}
+#endif
+
+gchar *atspi_dbus_name = NULL;
+static gboolean atspi_no_register = FALSE;
 
 static GOptionEntry atspi_option_entries[] =
 {
@@ -240,6 +309,12 @@ adaptor_init (gint *argc, gchar **argv[])
   DBusConnection *bus;
   AtkObject *root;
   gchar *introspection_directory;
+  static gboolean inited = FALSE;
+
+  if (inited)
+    return 0;
+
+  inited = TRUE;
 
   DRoutePath *treepath, *accpath;
 
@@ -260,18 +335,24 @@ adaptor_init (gint *argc, gchar **argv[])
 
   /* Set up D-Bus connection and register bus name */
   dbus_error_init (&error);
-  atk_adaptor_app_data->bus = dbus_bus_get (DBUS_BUS_SESSION, &error);
+  atk_adaptor_app_data->bus = spi_atk_bridge_get_bus ();
   if (!atk_adaptor_app_data->bus)
   {
-    g_warning ("AT-SPI Couldn't connect to D-Bus: %s\n", error.message);
     g_free(atk_adaptor_app_data);
     atk_adaptor_app_data = NULL;
     return 0;
   }
-  if (atspi_dbus_name != NULL &&
-      dbus_bus_request_name(atk_adaptor_app_data->bus, atspi_dbus_name, 0, &error))
+
+  if (atspi_dbus_name != NULL)
   {
-    g_print("AT-SPI Recieved D-Bus name - %s\n", atspi_dbus_name);
+    if (dbus_bus_request_name(atk_adaptor_app_data->bus, atspi_dbus_name, 0, &error))
+    {
+      g_print("AT-SPI Recieved D-Bus name - %s\n", atspi_dbus_name);
+    }
+    else
+    {
+      g_print("AT-SPI D-Bus name requested but could not be allocated - %s\n", atspi_dbus_name);
+    }
   }
 
   dbus_connection_setup_with_g_main(atk_adaptor_app_data->bus, g_main_context_default());
@@ -291,7 +372,7 @@ adaptor_init (gint *argc, gchar **argv[])
   accpath = droute_add_many (atk_adaptor_app_data->droute,
                              "/org/freedesktop/atspi/accessible",
                              NULL,
-                             (DRouteGetDatumFunction) atk_dbus_path_to_object);
+                             (DRouteGetDatumFunction) atk_dbus_path_to_gobject);
 
   /* Register all interfaces with droute and set up application accessible db */
   spi_initialize_tree (treepath);
@@ -317,8 +398,14 @@ adaptor_init (gint *argc, gchar **argv[])
   /* Register methods to send D-Bus signals on certain ATK events */
   spi_atk_register_event_listeners ();
 
+#ifdef __ATK_PLUG_H__
+  /* Hook our plug-and socket functions */
+  install_plug_hooks ();
+#endif
+
   /* Register this app by sending a signal out to AT-SPI registry daemon */
-  register_application (atk_adaptor_app_data);
+  if (!atspi_no_register)
+    register_application (atk_adaptor_app_data);
 
   g_atexit (exit_func);
 
@@ -339,4 +426,26 @@ gtk_module_init (gint *argc, gchar **argv[])
   return 0;
 }
 
+void
+gnome_accessibility_module_init (void)
+{
+  const gchar *load_bridge = g_getenv ("NO_AT_BRIDGE");
+
+  if (!load_bridge || g_ascii_strtod (load_bridge, NULL) == 0)
+    {
+      adaptor_init (NULL, NULL);
+
+      if (g_getenv ("AT_SPI_DEBUG"))
+        {
+	    g_print("Atk Accessibility bridge initialized\n");
+        }
+    }
+}
+
+void
+gnome_accessibility_module_shutdown (void)
+{
+  spi_atk_deregister_event_listeners ();
+  exit_func ();
+}
 /*END------------------------------------------------------------------------*/
